"""
 ***************************************************************************
                             Haar transform
                             -------------------
    begin                : Sat Dec 14 2002
    copyright            : (C) 2002 by Ricardo Niederberger Cabral
    email                : nieder@mail.ru
    Time-stamp:            <2003-01-07 23:17:39 rnc>
 ***************************************************************************

 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************

TODO:
- Try using the same query and standard transforms on another color space and then using a weighted sum as the final candidate score
"""
import sys,os,traceback,Error,time,bisect
from string import *
try:
    from qt import *
except:
    Error.qtImportError()
    
####### DEFINES
supported_ext=["jpg","gif","bmp","png","xbm","pnm"]       # thats what QT seems to support. gif and jpg may not be present. TODO: detect it
####### GLBOAL VARS
bin=[]

def initBin():
    """ initialize table telling which coeff bin to use for a given wavelet coef."""
    for i in range(128):
        for j in range(128):
            bin.append(min(max(i,j),5))
   
def cenRoot(i):
    return [i[0] /11.314,i[1] /11.314,i[2] /11.314]
def cenRoot2(i):
    return i/11.314
class sortpair:
    """ helper class used when sorting the list of top X matches"""
    def __init__(self, a,b):
        self.pair=[a,b]       
    def __lt__(self, other):        
        return self.pair[1]<other.pair[1]


class HaarEngine:
    def __init__(self,ndb):
        self.curdb=ndb
    def calcHaarSlow(self,aImage,M=40):
        #convert to yiq colorspace
        data=[]
        for i in range(128):
            for j in range(128):
                curpixel=[\
                     (qRed(aImage.pixel(j,i))*0.299+qGreen(aImage.pixel(j,i))*0.587+qBlue(aImage.pixel(j,i))*0.114)/256.0,\
                     (qRed(aImage.pixel(j,i))*0.596+qGreen(aImage.pixel(j,i))*(-0.274)+qBlue(aImage.pixel(j,i))*(-0.322))/256.0,\
                     (qRed(aImage.pixel(j,i))*0.212+qGreen(aImage.pixel(j,i))*(-0.523)+qBlue(aImage.pixel(j,i))*0.311)/256.0\
                     ]
                data.append(curpixel)
        #decompose rows
        for row in range(128):
            #A=A/root(h)
            h=128            
            data[row*128:(row+1)*128]=map(cenRoot ,data[row*128:(row+1)*128])
            while(h>1):
                h=h/2
                Ab=2*h*[0]
                for i in range(h):
                    Ab[i]=[(data[row*128+2*i][0]+data[row*128+2*i+1][0])/1.414,
                           (data[row*128+2*i][1]+data[row*128+2*i+1][1])/1.414,
                           (data[row*128+2*i][2]+data[row*128+2*i+1][2])/1.414]           
                    Ab[i+h]=[(data[row*128+2*i][0]-data[row*128+2*i+1][0])/1.414,(data[row*128+2*i][1]-data[row*128+2*i+1][1])/1.414,(data[row*128+2*i][2]-data[row*128+2*i+1][2])/1.414]
                data[row*128:row*128+2*h]=Ab
        #decompose cols
        for col in range(128):
            #A=A/root(h)
            h=128
            for w in range(128):
                data[w*128+col]=map(cenRoot2,data[w*128+col])
            while(h>1):
                h=h/2
                Ab=2*h*[0]
                for i in range(h):
                    Ab[i]=[(data[2*i*128+col][0]+data[(2*i+1)*128+col][0])/1.414,(data[2*i*128+col][1]+data[(2*i+1)*128+col][1])/1.414,(data[2*i*128+col][2]+data[(2*i+1)*128+col][2])/1.414]
                    Ab[i+h]=[(data[2*i*128+col][0]-data[(2*i+1)*128+col][0])/1.414,(data[2*i*128+col][1]-data[(2*i+1)*128+col][1])/1.414,(data[2*i*128+col][2]-data[(2*i+1)*128+col][2])/1.414]
                for w in range(2*h):
                    data[w*128+col]=Ab[w]
        avgl=data[0] # average luminance
        #get largest M coeffs
        best=[range(-9999,-9999+M),range(-9999,-9999+M),range(-9999,-9999+M)]
        for c in range(3):
            # get largest
            for i in range(len(data)):
                bisect.insort_left(best[c],abs(data[i][c]))
                del best[c][0]
        # truncate and quantize        
        data=map(lambda i: [(abs(i[0])>best[0][0])*i[0],(abs(i[1])>best[1][0])*i[1],(abs(i[2])>best[2][0])*i[2]],data)
        return (avgl,data)        
    def calcHaarFast(self,aImage,M=40):
        #convert to yiq colorspace
        cdata1=haar.new_darray(16384)
        cdata2=haar.new_darray(16384)
        cdata3=haar.new_darray(16384)

        cn=0
        for i in range(128):
            for j in range(128):
                haar.darray_set(cdata1,cn,qRed(aImage.pixel(j,i)))
                haar.darray_set(cdata2,cn,qGreen(aImage.pixel(j,i)))
                haar.darray_set(cdata3,cn,qBlue(aImage.pixel(j,i)))
                cn=cn+1
        #for i in range(128):
        #    haar.extractRGB(cdata1,cdata2,cdata3,i,aImage.bits(i))
            
        haar.transform(cdata1,cdata2,cdata3)
        avgl=[haar.darray_get(cdata1,0),haar.darray_get(cdata2,0),haar.darray_get(cdata3,0)] # average luminance
        #get largest M coeffs
        best=[range(-9999,-9999+M),range(-9999,-9999+M),range(-9999,-9999+M)]
        cadata1=haar.absarray(cdata1)
        cadata2=haar.absarray(cdata2)
        cadata3=haar.absarray(cdata3)
        # get largest
        # Performance changelog
        ## Adding took 41.857539 -now
        ## Adding took 23.867980 -without insort
        ## Adding took 12.337539 -without insort without rgb init        
        for i in range(16384):
            bisect.insort_left(best[0],haar.darray_get(cadata1,i))
            del best[0][0]
            bisect.insort_left(best[1],haar.darray_get(cadata2,i))
            del best[1][0]
            bisect.insort_left(best[2],haar.darray_get(cadata3,i))
            del best[2][0]
        # truncate and quantize
        haar.truncq(cadata1,cdata1,best[0][0])
        haar.truncq(cadata2,cdata2,best[1][0])
        haar.truncq(cadata3,cdata3,best[2][0])
        data=16384*[0,0,0]
        for i in range(16384):
            data[i]=[haar.darray_get(cdata1,i),haar.darray_get(cdata2,i),haar.darray_get(cdata3,i)]
        haar.del_darray(cadata1)
        haar.del_darray(cadata2)
        haar.del_darray(cadata3)
        haar.del_darray(cdata1)
        haar.del_darray(cdata2)
        haar.del_darray(cdata3)
        return (avgl,data)
    def query(self,aImage,numres=10,scanned=1):
        if scanned:
            w=[[5.00,19.21,34.37],[0.83,1.26,0.36],[1.01,0.44,0.45],[0.52,0.53,0.14],[0.47,0.28,0.18],[0.3,0.14,0.27]]
        else:
            w=[[4.04,15.14,22.62],[0.78,0.92,0.40],[0.46,0.53,0.63],[0.42,0.26,0.25],[0.41,0.14,0.15],[0.32,0.07,0.38]]
        sig=self.calcHaar(aImage)
        if not len(sig): #error opening file
            return []
        data=sig[1]
        score=len(self.curdb.img)*[0]
        keys=self.curdb.img.keys()
        for i in range(len(self.curdb.img)):
            score[i]=sortpair(i,0)
            for c in range(3):
                score[i].pair[1]=score[i].pair[1]+w[0][c]*abs(self.curdb.img[keys[i]][2][c]-sig[0][c])
        for i in range(16384):
            for c in range(3):                
                if data[i][c]:
                    pn=0
                    if data[i][c]<0: pn=1
                    for buck in self.curdb.buckets[i][c][pn]:
                        post=keys.index(buck)
                        score[post].pair[1]=score[post].pair[1]-w[bin[i]][c]
        #get best results
        best=[]
        # get largest
        numres=numres+1
        for i in range(len(self.curdb.img)):           
            bisect.insort_left(best,score[i])
            if len(best)>numres: del best[numres]
        best=best[1:]
        #set pair[0] to imgid
        for i in range(len(best)):    
            best[i].pair[0]=keys[best[i].pair[0]]
        return best
    def textquery(self,txt,numres=10):
        res=[]
        found=[]
        metaq=txt[0]
        logic=txt[1]
        pres=[]
        for fi in range(len(metaq)):
            pres.append([])
        for fil in self.curdb.meta.keys():            
            for fi in range(len(pres)):
                try:
                    field=metaq[fi][0]
                    if not self.curdb.meta[fil].has_key(field) or not self.curdb.meta[fil][field]:
                        continue
                    else:
                        if find(lower(self.curdb.meta[fil][field]),lower(metaq[fi][1])) != -1:
                            pres[fi].append(fil)
                            
                except:
                    traceback.print_exc()
                    pass
        if logic=="AND":
            min=0
            for fi in range(len(pres)):
                if len(pres[fi]) < len(pres[min]):
                    min=fi
            for id in pres[min]:
                failed=0
                for fi in range(len(pres)):
                    if id not in pres[fi]:                        
                        failed=1
                        break
                if not failed and id not in res: res.append(id)                
        elif logic=="OR":
            for fi in range(len(pres)):
                for id in pres[fi]:
                    if id not in res: res.append(id)
        fres=[]
        for id in res:
            fres.append(sortpair(id,0))
            if len(fres)>=numres:
                return fres
        return fres
try:
    import haar
    HaarEngine.calcHaar=HaarEngine.calcHaarFast
    print "Using fast Haar C extension."
except:
    print "Warning: Unable to load the C extension \"haar.so\" module. Make sure you installed imgSeek correctly, and mail any install bug to nieder@mail.ru.\nUsing the slow python based transform instead. (Everything will work as intended, but adding files will be slow)"
    HaarEngine.calcHaar=HaarEngine.calcHaarSlow

initBin()

